/*
 * Copyright (c) 2005 Silicon Graphics, Inc.
 * All rights reserved.
 *
 *   Michael A. Raymond <mraymond@sgi.com>
 */

/*
 *  This file is part of libuli.
 *
 *  libuli is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation; either version 2.1 of the License, or
 *  (at your option) any later version.
 *
 *  libuli is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the Lesser GNU General Public License
 *  along with libuli; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 */

#include <stdio.h>
#include <errno.h>
#include <fcntl.h>
#include <stdlib.h>
#include <uli.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <asm/bitops.h>

/*
 *   This is the library for User Level Interrupt (ULI) functionality.  ULIs
 * are a feature wherein hardware interrupts can be handled by user code in
 * user space quickly and safely.  They are useful for programs doing
 * memory mapped I/O such as embedded systems, HWIL simulators, and other
 * high performance systems.
 *
 *   This library exists to make it easy for user programs to use the ULI
 * functionality.  It provides interface routines so that programs can
 * register interrupt handlers for different types of hardware interrupts.
 * It also gives routines to block and unblock interrupt handling, and for
 * suspending thread execution until the next interrupt is handled.
 *
 *   The kernel level details of how ULIs are implemented are presented in
 * the kernel source files uli.c and uli.h.  From the user's prospective,
 * ULIs may be handled at any time, act like a separate thread within a
 * process, and execute with their own stacks.  They are disimilar from
 * POSIX signals in that their "like a separate thread" characteristic means
 * that even a single threaded process may not necessarily be preempted by
 * handler execution.
 *
 *   To make things simpler for the kernel, it is required that ULI handlers
 * not cause any TLB misses that require pages to be fetched from storage to
 * be handled.  This prevents the handler from having to block during the
 * fetch.  To achieve this requirement it is highly recomended that processes
 * using ULIs lock all of their memory into core.  If a handler does cause
 * a page fault then its process will be killed.  Note that a common cause
 * of fatal page faults occurs when handlers call a previously unused library
 * routine which results in a fault on the backing store of a jump table.
 *
 *                                                  -------
 *      --------         ------------------------  | User  |
 *     |Shared  |       | User        -------- <-|-|Process|
 *     |Memory /|       |Process <-->|  ULI   |  |  -------
 *     |Mapped  |<------|----------->|Function|  |      -------
 *     |Memory  |       |             -------- <-|-----| User  |
 *      --------        |Memory Access   A       |     |Process|
 *                       ------------------------       -------
 * User Level                            |
 * --------------------------------------------------------------------
 * Interrupt Level                       |
 *                                     ------
 *                                    |Kernel|
 *                                    |ISR   | <--- Interrupt
 *                                     ------
 */

/* These are the legal flag values for the callback structure's lock */
#define LOCK_CLEARED	0x00	/* Unlocked with no deferred interrupts */
#define LOCK_HELD	0x01	/* Lock held and no interrupts will be processed */
#define LOCK_INTR_PENDING	0x10	/* An interrupt has come in */

#define STACK_SIZE 1024	/* Default stack size */
#define MIN_STACK 1024	/* Minimum legal stack size */

/*
 * This user space argument structure exists one per registered
 * ULI.  The initial ULI handler is passed its location and
 * uses it to call the user's handler.
 */
struct callback {
	void (*func)(void*);	/* User's handler function */
	void *arg;		/* Argument to user's handler function */
	volatile int lock;	/* Lock to synchronize blocking */
	int fd;		/* File descriptor of open ULI device */
	char * stack;		/* The stack we allocated for the user */
};

#define CALLBACK(vp) ((struct callback*)(vp))

/*
 * On IA64 function pointers point to a two element structure
 * containing the actual function pointer and the global pointer
 * to use with it.  We need to pass both to the kernel ULI code,
 * so we use this structure to extract that information.
 */
struct indir {
	void (*hfunc)(struct callback*);	/* Actual handler function */
	caddr_t gp;	/* GP within handler */
};

/* Architecture specific synchronization functions */
extern int ULI_tao(volatile int *, int);
extern int ULI_tas(volatile int *, int);

/*
 * This is the initial handler code that the kernel calls into.
 * From here we call into the user's handler.
 */
static void
ULI_callback(struct callback * arg)
{
	/*
	 * Try to acquire the interrupt lock for this intr. This code runs at
	 * interrupt level, so we can't spin waiting for the lock. If we get the
	 * lock, we can just process the interrupt and return. Otherwise we do
	 * nothing, but we do leave the INTR_PENDING bit set in the lock to
	 * indicate that an interrupt occurred while blocked and it should be
	 * processed later when the top half unblocks the interrupt.
	 */
	if (ULI_tao(&arg->lock, LOCK_INTR_PENDING) == 0) {
		arg->func(arg->arg);

		/* Unlock */
		arg->lock = LOCK_CLEARED;
	}

	/* Use some method to return to the kernel */
	write(arg->fd, (char*)ULI_RETURN_ADDR, 100);

	/*NOTREACHED*/
	return;
}

/*
 * Block execution of the handler for this ULI.  What we really do
 * is mark its callback structure so that when the interrupt gets handled
 * by ULI_callback it just returns.  If the ULI is currently being
 * handled on another CPU we spin until it completes.
 */
void
ULI_block_intr(void * intr)
{
	/* This is called from regular code, so it's OK to spin */
	while (ULI_tao(&CALLBACK(intr)->lock, LOCK_HELD)) {
	}
}

/*
 * Unblock the interrupt and check if there is a deferred interrupt.
 * We may be acting on the lock at the same time as the ULI handler
 * and another CPU, so we may need to loop a few times to sync up
 * with them properly.  We must check for a deferred interrupt
 * before processing it, we must process it before unblocking the
 * interrupt, and we must unblock the interrupt before checking for
 * a deferred interrupt.
 */
void
ULI_unblock_intr(void * intr)
{
	/* Unblock the interrupt and check if there was a deferred interrupt */
	while (ULI_tas(&CALLBACK(intr)->lock, LOCK_CLEARED) &
		LOCK_INTR_PENDING) {
		int lock;
		/*
		 * There was a deferred interrupt, so we must reacquire
		 * the lock to process it.  Keep looping while LOCK_HELD
		 * was set by another CPU so that we don't cancel their
		 * blocking operation.
		 */
		while ((lock = ULI_tao(&CALLBACK(intr)->lock, LOCK_HELD)) &
			LOCK_HELD) {
		}

		/*
		 * If LOCK_INTR_PENDING was set before we reacquirred the lock
		 * then the handler has come in and is currently handling
		 * the interrupt.  There's no reason for us to handle it
		 * again, and the handler will clear the lock when it's done,
		 * so we just exit.
		 */
		if (lock & LOCK_INTR_PENDING) {
			break;
		}
		
		/* Execute the deferred interrupt */
		CALLBACK(intr)->func(CALLBACK(intr)->arg);
	}
}

/*
 * Check the values the user requested for an interrupt stack.
 * If none was requested then we allocate one.  We let the
 * kernel handle properly aligning things.  This gives the side
 * benefit of allowing us to free it later.
 */
static int
getstack(struct uliargs * uargs)
{
	char * newstack;
	char * stack = uargs->sp;

	/* If caller specfied a stack */
	if (uargs->sp) {
		/* Pass the top of the stack */
		uargs->sp = uargs->sp + uargs->stacksize;

		if (uargs->sp <= stack)
			return EINVAL;

		return 0;
	}

	/* We are expected to allocate a stack */
	if (uargs->stacksize == 0) {
		uargs->stacksize = STACK_SIZE;
	}
	if (uargs->stacksize < MIN_STACK) {
		return EINVAL;
	}

	if ((newstack = (char*)malloc(uargs->stacksize)) == NULL) {
		return ENOMEM;
	}

	/* Pass the kernel the top of the stack */
	uargs->sp = newstack + uargs->stacksize;
	return 0;
}

/*
 * Register a ULI handler for a particular (IRQ,CPU) pair.
 */
void *
ULI_register_irq(short cpu, void (*func)(void*), void * arg,
		   unsigned int nsemas, char * stack, int stacksize, int line)
{
	int file, res;
	struct uliargs uargs;
	struct callback * newcallback;
	struct indir * indir;

	/* Allocate a per-ULI control structure */
	if ((newcallback = (struct callback*)malloc(sizeof(struct callback)))
	    == NULL) {
		errno = ENOMEM;
		return NULL;
	}

	uargs.sp = stack;
	uargs.stacksize = stacksize;

	/* Set up the interrupt stack */
	if ((res = getstack(&uargs))) {
		free(newcallback);
		errno = res;
		return NULL;
	}

	/* Mark if we did the stack allocation */
	if (stack == NULL) {
		newcallback->stack = uargs.sp - uargs.stacksize;
	}

	/* Open the ULI device so we can communicate with the kernel */
	if ((file = open(ULI_DEV, O_RDWR)) < 0) {
		if (newcallback->stack) {
			free(newcallback->stack);
		}
		free(newcallback);
		return NULL;
	}

	/* Close the ULI file on exec */
	fcntl(file, F_SETFD, FD_CLOEXEC);

	/* Preresolve "write" to save time in the ULI handler */
	write(file, (char*)ULI_RESOLV_ADDR, 100);

	/*
	 * Get the PC/GP pair for ULI_callback.  This is architecture
	 * specific.
	 */
	indir = (struct indir*)ULI_callback;
	uargs.pc = (caddr_t)indir->hfunc;
	uargs.gp = indir->gp;

	/* Set up the arguments to be passed to the kernel */
	uargs.funcarg = newcallback;
	uargs.id = line;
	uargs.nsemas = nsemas;
	uargs.intarg = cpu;

	newcallback->func = func;
	newcallback->arg = arg;
	newcallback->fd = file;

	/*
	 * Don't allow an interrupt until we've set everything up.
	 * Preresolve ULI_tao at the same time.
	 */
	newcallback->lock = LOCK_CLEARED;
	ULI_tao(&newcallback->lock, LOCK_HELD);

	/* Attempt to register the ULI */
	if (ioctl(file, IOC_ULI_REG_IRQ, &uargs) < 0) {
		if (newcallback->stack) {
			free(newcallback->stack);
		}
		free(newcallback);
		close(file);
		return NULL;
	}

	newcallback->lock = LOCK_CLEARED;

	return newcallback;
}

/*
 * Destroy a particular ULI and free its resources.
 */
int
ULI_destroy(void * intr)
{
	/* Destroy the kernel side state */
	if (close(CALLBACK(intr)->fd)) {
		return -1;
	}

	/* Free the stack if we allocated one */
	if (CALLBACK(intr)->stack) {
		free(CALLBACK(intr)->stack);
	}

	/* Free the control structure */
	free(CALLBACK(intr));

	return 0;
}

/*
 * Block on a specific semaphore for the passed ULI.
 */
int
ULI_sleep(void * intr, int sema)
{
	if (write(CALLBACK(intr)->fd, (char*)ULI_SLEEP_ADDR, sema) < 0) {
		errno = EINVAL;
		return -1;
	}

	return 0;
}

/*
 * Wakeup a blocker on a specific semaphore for the passed ULI.
 */
int
ULI_wakeup(void * intr, int sema)
{
	if (!intr) {
		return -1;
	}

	if (write(CALLBACK(intr)->fd, (char*)ULI_WAKEUP_ADDR, sema) < 0) {
		errno = EINVAL;
		return -1;
	}

	return 0;
}
